from __future__ import annotations
from typing import Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from telebox.dispatcher.dispatcher import Dispatcher
from telebox.dispatcher.handlers.event import AbstractEventHandler
from telebox.dispatcher.filters.event_filter import AbstractEventBaseFilter
from telebox.dispatcher.utils.rate_limiter.rate_limit import RateLimit
from telebox.utils.not_set import NotSet, NOT_SET


class Router:

    def __init__(
        self,
        dispatcher: Dispatcher,
        filter_: Optional[AbstractEventBaseFilter] = None
    ):
        self._dispatcher = dispatcher
        self._filter = filter_

    def __add__(self, other) -> "Router":
        if isinstance(other, AbstractEventBaseFilter):
            return Router(self._dispatcher, self._get_filter(other))

        return NotImplemented

    def __iadd__(self, other) -> "Router":
        return self.__add__(other)

    def add_message_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._dispatcher.add_message_handler(
            handler,
            self._get_filter(filter_),
            rate_limit
        )

    def add_edited_message_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._dispatcher.add_edited_message_handler(
            handler,
            self._get_filter(filter_),
            rate_limit
        )

    def add_channel_post_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._dispatcher.add_channel_post_handler(
            handler,
            self._get_filter(filter_),
            rate_limit
        )

    def add_edited_channel_post_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._dispatcher.add_edited_channel_post_handler(
            handler,
            self._get_filter(filter_),
            rate_limit
        )

    def add_media_group_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._dispatcher.add_media_group_handler(
            handler,
            self._get_filter(filter_),
            rate_limit
        )

    def add_inline_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._dispatcher.add_inline_query_handler(
            handler,
            self._get_filter(filter_),
            rate_limit
        )

    def add_chosen_inline_result_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._dispatcher.add_chosen_inline_result_handler(
            handler,
            self._get_filter(filter_),
            rate_limit
        )

    def add_callback_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._dispatcher.add_callback_query_handler(
            handler,
            self._get_filter(filter_),
            rate_limit
        )

    def add_shipping_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._dispatcher.add_shipping_query_handler(
            handler,
            self._get_filter(filter_),
            rate_limit
        )

    def add_pre_checkout_query_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None,
        rate_limit: Union[RateLimit, None, NotSet] = NOT_SET
    ) -> None:
        self._dispatcher.add_pre_checkout_query_handler(
            handler,
            self._get_filter(filter_),
            rate_limit
        )

    def add_poll_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._dispatcher.add_poll_handler(handler, self._get_filter(filter_))

    def add_poll_answer_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._dispatcher.add_poll_answer_handler(handler, self._get_filter(filter_))

    def add_my_chat_member_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._dispatcher.add_my_chat_member_handler(handler, self._get_filter(filter_))

    def add_chat_member_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._dispatcher.add_chat_member_handler(handler, self._get_filter(filter_))

    def add_chat_join_request_handler(
        self,
        handler: AbstractEventHandler,
        filter_: Optional[AbstractEventBaseFilter] = None
    ) -> None:
        self._dispatcher.add_chat_join_request_handler(handler, self._get_filter(filter_))

    def _get_filter(self, filter_: AbstractEventBaseFilter) -> Optional[AbstractEventBaseFilter]:
        if filter_ is None:
            return self._filter
        elif self._filter is None:
            return filter_

        return self._filter & filter_



from typing import Optional, Any, Iterable
from urllib.parse import urlencode

from telebox.utils.not_set import NOT_SET


def get_username_link(username: str, *, with_tg: bool = False) -> str:
    if with_tg:
        return _get_parametrized_link(
            "tg://resolve",
            parameters={
                "domain": username
            }
        )
    else:
        return f"t.me/{username}"


def get_phone_number_link(number: str, *, with_tg: bool = False) -> str:
    number = number.lstrip("+")

    if with_tg:
        return _get_parametrized_link(
            "tg://resolve",
            parameters={
                "phone": number
            }
        )
    else:
        return f"t.me/+{number}"


def get_chat_invite_link(hash_: str, *, with_tg: bool = False) -> str:
    if with_tg:
        return _get_parametrized_link(
            "tg://join",
            parameters={
                "invite": hash_
            }
        )
    else:
        return f"t.me/+{hash_}"


def get_message_public_link(username: str, message_id: int, *, with_tg: bool = False) -> str:
    if with_tg:
        return _get_parametrized_link(
            "tg://resolve",
            parameters={
                "domain": username,
                "post": message_id
            }
        )
    else:
        return f"t.me/{username}/{message_id}"


def get_message_private_link(chat_id: int, message_id: int, *, with_tg: bool = False) -> str:
    if with_tg:
        return _get_parametrized_link(
            "tg://privatepost",
            parameters={
                "channel": chat_id,
                "post": message_id
            }
        )
    else:
        return f"t.me/c/{chat_id}/{message_id}"


def get_share_link(url: str, text: Optional[str] = None, *, with_tg: bool = False) -> str:
    return _get_parametrized_link(
        "tg://msg_url" if with_tg else "t.me/share",
        parameters={
            "url": url,
            "text": text
        }
    )


def get_video_chat_link(
    username: str,
    invite_hash: Optional[str] = None,
    *,
    with_tg: bool = False
) -> str:
    video_chat_value = NOT_SET if invite_hash is None else invite_hash

    if with_tg:
        return _get_parametrized_link(
            "tg://resolve",
            parameters={
                "domain": username,
                "videochat": video_chat_value
            }
        )
    else:
        return _get_parametrized_link(
            f"t.me/{username}",
            parameters={
                "videochat": video_chat_value
            }
        )


def get_livestream_link(
    username: str,
    invite_hash: Optional[str] = None,
    *,
    with_tg: bool = False
) -> str:
    livestream_value = NOT_SET if invite_hash is None else invite_hash

    if with_tg:
        return _get_parametrized_link(
            "tg://resolve",
            parameters={
                "domain": username,
                "livestream": livestream_value
            }
        )
    else:
        return _get_parametrized_link(
            f"t.me/{username}",
            parameters={
                "livestream": livestream_value
            }
        )


def get_voice_chat_link(
    username: str,
    invite_hash: Optional[str] = None,
    *,
    with_tg: bool = False
) -> str:
    voice_chat_value = NOT_SET if invite_hash is None else invite_hash

    if with_tg:
        return _get_parametrized_link(
            "tg://resolve",
            parameters={
                "domain": username,
                "voicechat": voice_chat_value
            }
        )
    else:
        return _get_parametrized_link(
            f"t.me/{username}",
            parameters={
                "voicechat": voice_chat_value
            }
        )


def get_add_stickers_link(name: str, *, with_tg: bool = False) -> str:
    if with_tg:
        return _get_parametrized_link(
            "tg://addstickers",
            parameters={
                "set": name
            }
        )
    else:
        return f"t.me/addstickers/{name}"


def get_add_emoji_link(name: str, *, with_tg: bool = False) -> str:
    if with_tg:
        return _get_parametrized_link(
            "tg://addemoji",
            parameters={
                "set": name
            }
        )
    else:
        return f"t.me/addemoji/{name}"


def get_mtproxy_link(host: str, port: int, secret: str, *, with_tg: bool = False) -> str:
    return _get_parametrized_link(
        "tg://proxy" if with_tg else "t.me/proxy",
        parameters={
            "server": host,
            "port": port,
            "secret": secret
        }
    )


def get_socks5_proxy_link(
    host: str,
    port: int,
    username: Optional[str] = None,
    password: Optional[str] = None,
    *,
    with_tg: bool = False
) -> str:
    return _get_parametrized_link(
        "tg://socks" if with_tg else "t.me/socks",
        parameters={
            "server": host,
            "port": port,
            "user": username,
            "pass": password
        }
    )


def get_add_theme_link(name: str, *, with_tg: bool = False) -> str:
    if with_tg:
        return _get_parametrized_link(
            "tg://addtheme",
            parameters={
                "slug": name
            }
        )
    else:
        return f"t.me/addtheme/{name}"


def get_bot_link(username: str, payload: str, *, with_tg: bool = False) -> str:
    if with_tg:
        return _get_parametrized_link(
            "tg://resolve",
            parameters={
                "domain": username,
                "start": payload
            }
        )
    else:
        return _get_parametrized_link(
            f"t.me/{username}",
            parameters={
                "start": payload
            }
        )


def get_group_bot_link(
    username: str,
    payload: Optional[str] = None,
    admin_rights: Optional[Iterable[str]] = None,
    *,
    with_tg: bool = False
) -> str:
    payload_value = NOT_SET if payload is None else payload
    admin_value = "+".join(admin_rights) if admin_rights is not None else None

    if with_tg:
        return _get_parametrized_link(
            "tg://resolve",
            parameters={
                "domain": username,
                "startgroup": payload_value,
                "admin": admin_value
            }
        )
    else:
        return _get_parametrized_link(
            f"t.me/{username}",
            parameters={
                "startgroup": payload_value,
                "admin": admin_value
            }
        )


def get_channel_bot_link(
    username: str,
    admin_rights: Iterable[str],
    *,
    with_tg: bool = False
) -> str:
    admin_value = "+".join(admin_rights)

    if with_tg:
        return _get_parametrized_link(
            "tg://resolve",
            parameters={
                "domain": username,
                "startchannel": NOT_SET,
                "admin": admin_value
            }
        )
    else:
        return _get_parametrized_link(
            f"t.me/{username}",
            parameters={
                "startchannel": NOT_SET,
                "admin": admin_value
            }
        )


def get_game_link(username: str, name: str, *, with_tg: bool = False) -> str:
    if with_tg:
        return _get_parametrized_link(
            "tg://resolve",
            parameters={
                "domain": username,
                "game": name
            }
        )
    else:
        return _get_parametrized_link(
            f"t.me/{username}",
            parameters={
                "game": name
            }
        )


def get_user_link(id_: int) -> str:
    return _get_parametrized_link(
        "tg://user",
        parameters={
            "id": id_
        }
    )


def _get_parametrized_link(uri: str, parameters: dict[str, Any]) -> str:
    value_parameters = {}
    no_value_parameters = []

    for name, value in parameters.items():
        if value is NOT_SET:
            no_value_parameters.append(name)
        elif value is not None:
            value_parameters[name] = value

    link = uri

    if value_parameters or no_value_parameters:
        link += "?"

        if value_parameters:
            link += urlencode(value_parameters, encoding="UTF-8")

            if no_value_parameters:
                link += "&" + "&".join(no_value_parameters)
        elif no_value_parameters:
            link += "&".join(no_value_parameters)

    return link



from dataclasses import dataclass
from typing import Optional, Union

from telebox.bot.enums.update_content_type import UpdateContentType
from telebox.bot.errors import UnknownUpdateContentError
from telebox.bot.types.type import Type
from telebox.bot.types.types.message import Message
from telebox.bot.types.types.inline_query import InlineQuery
from telebox.bot.types.types.chosen_inline_result import ChosenInlineResult
from telebox.bot.types.types.callback_query import CallbackQuery
from telebox.bot.types.types.shipping_query import ShippingQuery
from telebox.bot.types.types.pre_checkout_query import PreCheckoutQuery
from telebox.bot.types.types.poll import Poll
from telebox.bot.types.types.poll_answer import PollAnswer
from telebox.bot.types.types.chat_member_updated import ChatMemberUpdated
from telebox.bot.types.types.chat_join_request import ChatJoinRequest


UpdateContent = Union[
    Message,
    InlineQuery,
    ChosenInlineResult,
    CallbackQuery,
    ShippingQuery,
    PreCheckoutQuery,
    Poll,
    PollAnswer,
    ChatMemberUpdated,
    ChatJoinRequest
]


@dataclass(unsafe_hash=True)
class Update(Type):
    update_id: int
    message: Optional[Message] = None
    edited_message: Optional[Message] = None
    channel_post: Optional[Message] = None
    edited_channel_post: Optional[Message] = None
    inline_query: Optional[InlineQuery] = None
    chosen_inline_result: Optional[ChosenInlineResult] = None
    callback_query: Optional[CallbackQuery] = None
    shipping_query: Optional[ShippingQuery] = None
    pre_checkout_query: Optional[PreCheckoutQuery] = None
    poll: Optional[Poll] = None
    poll_answer: Optional[PollAnswer] = None
    my_chat_member: Optional[ChatMemberUpdated] = None
    chat_member: Optional[ChatMemberUpdated] = None
    chat_join_request: Optional[ChatJoinRequest] = None

    @property
    def content(self) -> tuple[UpdateContent, UpdateContentType]:
        if self.message is not None:
            return self.message, UpdateContentType.MESSAGE
        elif self.edited_message is not None:
            return self.edited_message, UpdateContentType.EDITED_MESSAGE
        elif self.channel_post is not None:
            return self.channel_post, UpdateContentType.CHANNEL_POST
        elif self.edited_channel_post is not None:
            return self.edited_channel_post, UpdateContentType.EDITED_CHANNEL_POST
        elif self.inline_query is not None:
            return self.inline_query, UpdateContentType.INLINE_QUERY
        elif self.chosen_inline_result is not None:
            return self.chosen_inline_result, UpdateContentType.CHOSEN_INLINE_RESULT
        elif self.callback_query is not None:
            return self.callback_query, UpdateContentType.CALLBACK_QUERY
        elif self.shipping_query is not None:
            return self.shipping_query, UpdateContentType.SHIPPING_QUERY
        elif self.pre_checkout_query is not None:
            return self.pre_checkout_query, UpdateContentType.PRE_CHECKOUT_QUERY
        elif self.poll is not None:
            return self.poll, UpdateContentType.POLL
        elif self.poll_answer is not None:
            return self.poll_answer, UpdateContentType.POLL_ANSWER
        elif self.my_chat_member is not None:
            return self.my_chat_member, UpdateContentType.MY_CHAT_MEMBER
        elif self.chat_member is not None:
            return self.chat_member, UpdateContentType.CHAT_MEMBER
        elif self.chat_join_request is not None:
            return self.chat_join_request, UpdateContentType.CHAT_JOIN_REQUEST

        raise UnknownUpdateContentError("Unknown update content!")



import html
import re


_markdown_pattern = re.compile(r"([_*\[\]()~`>#+\-=|{}.!\\])")


def get_escaped_html_text(text: str) -> str:
    return html.escape(text, quote=False)


def get_escaped_markdown_text(text: str) -> str:
    return re.sub(_markdown_pattern, r"\\\1", text)



from dataclasses import dataclass

from telebox.bot.types.type import Type


@dataclass(unsafe_hash=True)
class Invoice(Type):
    title: str
    description: str
    start_parameter: str
    currency: str
    total_amount: int



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type
from telebox.bot.types.types.keyboard_button import KeyboardButton


@dataclass(unsafe_hash=True)
class ReplyKeyboardMarkup(Type):
    keyboard: list[list[KeyboardButton]]
    resize_keyboard: Optional[bool] = None
    one_time_keyboard: Optional[bool] = None
    input_field_placeholder: Optional[str] = None
    selective: Optional[bool] = None



from .update import Update
from .webhook_info import WebhookInfo
from .user import User
from .chat import Chat
from .message import Message
from .message_id import MessageId
from .message_entity import MessageEntity
from .photo_size import PhotoSize
from .animation import Animation
from .audio import Audio
from .document import Document
from .video import Video
from .video_note import VideoNote
from .voice import Voice
from .contact import Contact
from .dice import Dice
from .poll_option import PollOption
from .poll_answer import PollAnswer
from .poll import Poll
from .location import Location
from .venue import Venue
from .web_app_data import WebAppData
from .proximity_alert_triggered import ProximityAlertTriggered
from .message_auto_delete_timer_changed import MessageAutoDeleteTimerChanged
from .forum_topic_created import ForumTopicCreated
from .forum_topic_closed import ForumTopicClosed
from .forum_topic_reopened import ForumTopicReopened
from .video_chat_scheduled import VideoChatScheduled
from .video_chat_started import VideoChatStarted
from .video_chat_ended import VideoChatEnded
from .video_chat_participants_invited import VideoChatParticipantsInvited
from .user_profile_photos import UserProfilePhotos
from .file import File
from .web_app_info import WebAppInfo
from .reply_keyboard_markup import ReplyKeyboardMarkup
from .keyboard_button import KeyboardButton
from .keyboard_button_poll_type import KeyboardButtonPollType
from .reply_keyboard_remove import ReplyKeyboardRemove
from .inline_keyboard_markup import InlineKeyboardMarkup
from .inline_keyboard_button import InlineKeyboardButton
from .login_url import LoginUrl
from .callback_query import CallbackQuery
from .force_reply import ForceReply
from .chat_photo import ChatPhoto
from .chat_invite_link import ChatInviteLink
from .chat_administrator_rights import ChatAdministratorRights
from .chat_member import ChatMember
from .chat_member_owner import ChatMemberOwner
from .chat_member_administrator import ChatMemberAdministrator
from .chat_member_member import ChatMemberMember
from .chat_member_restricted import ChatMemberRestricted
from .chat_member_left import ChatMemberLeft
from .chat_member_banned import ChatMemberBanned
from .chat_member_updated import ChatMemberUpdated
from .chat_join_request import ChatJoinRequest
from .chat_permissions import ChatPermissions
from .chat_location import ChatLocation
from .forum_topic import ForumTopic
from .bot_command import BotCommand
from .bot_command_scope import BotCommandScope
from .bot_command_scope_default import BotCommandScopeDefault
from .bot_command_scope_all_private_chats import BotCommandScopeAllPrivateChats
from .bot_command_scope_all_group_chats import BotCommandScopeAllGroupChats
from .bot_command_scope_all_chat_administrators import BotCommandScopeAllChatAdministrators
from .bot_command_scope_chat import BotCommandScopeChat
from .bot_command_scope_chat_administrators import BotCommandScopeChatAdministrators
from .bot_command_scope_chat_member import BotCommandScopeChatMember
from .menu_button import MenuButton
from .menu_button_commands import MenuButtonCommands
from .menu_button_web_app import MenuButtonWebApp
from .menu_button_default import MenuButtonDefault
from .response_parameters import ResponseParameters
from .input_media import InputMedia
from .input_media_photo import InputMediaPhoto
from .input_media_video import InputMediaVideo
from .input_media_animation import InputMediaAnimation
from .input_media_audio import InputMediaAudio
from .input_media_document import InputMediaDocument
from .input_file import InputFile
from .sticker import Sticker
from .sticker_set import StickerSet
from .mask_position import MaskPosition
from .inline_query import InlineQuery
from .inline_query_result import InlineQueryResult
from .inline_query_result_article import InlineQueryResultArticle
from .inline_query_result_photo import InlineQueryResultPhoto
from .inline_query_result_gif import InlineQueryResultGif
from .inline_query_result_mpeg4_gif import InlineQueryResultMpeg4Gif
from .inline_query_result_video import InlineQueryResultVideo
from .inline_query_result_audio import InlineQueryResultAudio
from .inline_query_result_voice import InlineQueryResultVoice
from .inline_query_result_document import InlineQueryResultDocument
from .inline_query_result_location import InlineQueryResultLocation
from .inline_query_result_venue import InlineQueryResultVenue
from .inline_query_result_contact import InlineQueryResultContact
from .inline_query_result_game import InlineQueryResultGame
from .inline_query_result_cached_photo import InlineQueryResultCachedPhoto
from .inline_query_result_cached_gif import InlineQueryResultCachedGif
from .inline_query_result_cached_mpeg4_gif import InlineQueryResultCachedMpeg4Gif
from .inline_query_result_cached_sticker import InlineQueryResultCachedSticker
from .inline_query_result_cached_document import InlineQueryResultCachedDocument
from .inline_query_result_cached_video import InlineQueryResultCachedVideo
from .inline_query_result_cached_voice import InlineQueryResultCachedVoice
from .inline_query_result_cached_audio import InlineQueryResultCachedAudio
from .input_message_content import InputMessageContent
from .input_text_message_content import InputTextMessageContent
from .input_location_message_content import InputLocationMessageContent
from .input_venue_message_content import InputVenueMessageContent
from .input_contact_message_content import InputContactMessageContent
from .input_invoice_message_content import InputInvoiceMessageContent
from .chosen_inline_result import ChosenInlineResult
from .sent_web_app_message import SentWebAppMessage
from .labeled_price import LabeledPrice
from .invoice import Invoice
from .shipping_address import ShippingAddress
from .order_info import OrderInfo
from .shipping_option import ShippingOption
from .successful_payment import SuccessfulPayment
from .shipping_query import ShippingQuery
from .pre_checkout_query import PreCheckoutQuery
from .passport_data import PassportData
from .passport_file import PassportFile
from .encrypted_passport_element import EncryptedPassportElement
from .encrypted_credentials import EncryptedCredentials
from .passport_element_error import PassportElementError
from .passport_element_error_data_field import PassportElementErrorDataField
from .passport_element_error_front_side import PassportElementErrorFrontSide
from .passport_element_error_reverse_side import PassportElementErrorReverseSide
from .passport_element_error_selfie import PassportElementErrorSelfie
from .passport_element_error_file import PassportElementErrorFile
from .passport_element_error_files import PassportElementErrorFiles
from .passport_element_error_translation_file import PassportElementErrorTranslationFile
from .passport_element_error_translation_files import PassportElementErrorTranslationFiles
from .passport_element_error_unspecified import PassportElementErrorUnspecified
from .game import Game
from .callback_game import CallbackGame
from .game_high_score import GameHighScore


__all__ = [
    "Update",
    "WebhookInfo",
    "User",
    "Chat",
    "Message",
    "MessageId",
    "MessageEntity",
    "PhotoSize",
    "Animation",
    "Audio",
    "Document",
    "Video",
    "VideoNote",
    "Voice",
    "Contact",
    "Dice",
    "PollOption",
    "PollAnswer",
    "Poll",
    "Location",
    "Venue",
    "WebAppData",
    "ProximityAlertTriggered",
    "MessageAutoDeleteTimerChanged",
    "ForumTopicCreated",
    "ForumTopicClosed",
    "ForumTopicReopened",
    "VideoChatScheduled",
    "VideoChatStarted",
    "VideoChatEnded",
    "VideoChatParticipantsInvited",
    "UserProfilePhotos",
    "File",
    "WebAppInfo",
    "ReplyKeyboardMarkup",
    "KeyboardButton",
    "KeyboardButtonPollType",
    "ReplyKeyboardRemove",
    "InlineKeyboardMarkup",
    "InlineKeyboardButton",
    "LoginUrl",
    "CallbackQuery",
    "ForceReply",
    "ChatPhoto",
    "ChatInviteLink",
    "ChatAdministratorRights",
    "ChatMember",
    "ChatMemberOwner",
    "ChatMemberAdministrator",
    "ChatMemberMember",
    "ChatMemberRestricted",
    "ChatMemberLeft",
    "ChatMemberBanned",
    "ChatMemberUpdated",
    "ChatJoinRequest",
    "ChatPermissions",
    "ChatLocation",
    "ForumTopic",
    "BotCommand",
    "BotCommandScope",
    "BotCommandScopeDefault",
    "BotCommandScopeAllPrivateChats",
    "BotCommandScopeAllGroupChats",
    "BotCommandScopeAllChatAdministrators",
    "BotCommandScopeChat",
    "BotCommandScopeChatAdministrators",
    "BotCommandScopeChatMember",
    "MenuButton",
    "MenuButtonCommands",
    "MenuButtonWebApp",
    "MenuButtonDefault",
    "ResponseParameters",
    "InputMedia",
    "InputMediaPhoto",
    "InputMediaVideo",
    "InputMediaAnimation",
    "InputMediaAudio",
    "InputMediaDocument",
    "InputFile",
    "Sticker",
    "StickerSet",
    "MaskPosition",
    "InlineQuery",
    "InlineQueryResult",
    "InlineQueryResultArticle",
    "InlineQueryResultPhoto",
    "InlineQueryResultGif",
    "InlineQueryResultMpeg4Gif",
    "InlineQueryResultVideo",
    "InlineQueryResultAudio",
    "InlineQueryResultVoice",
    "InlineQueryResultDocument",
    "InlineQueryResultLocation",
    "InlineQueryResultVenue",
    "InlineQueryResultContact",
    "InlineQueryResultGame",
    "InlineQueryResultCachedPhoto",
    "InlineQueryResultCachedGif",
    "InlineQueryResultCachedMpeg4Gif",
    "InlineQueryResultCachedSticker",
    "InlineQueryResultCachedDocument",
    "InlineQueryResultCachedVideo",
    "InlineQueryResultCachedVoice",
    "InlineQueryResultCachedAudio",
    "InputMessageContent",
    "InputTextMessageContent",
    "InputLocationMessageContent",
    "InputVenueMessageContent",
    "InputContactMessageContent",
    "InputInvoiceMessageContent",
    "ChosenInlineResult",
    "SentWebAppMessage",
    "LabeledPrice",
    "Invoice",
    "ShippingAddress",
    "OrderInfo",
    "ShippingOption",
    "SuccessfulPayment",
    "ShippingQuery",
    "PreCheckoutQuery",
    "PassportData",
    "PassportFile",
    "EncryptedPassportElement",
    "EncryptedCredentials",
    "PassportElementError",
    "PassportElementErrorDataField",
    "PassportElementErrorFrontSide",
    "PassportElementErrorReverseSide",
    "PassportElementErrorSelfie",
    "PassportElementErrorFile",
    "PassportElementErrorFiles",
    "PassportElementErrorTranslationFile",
    "PassportElementErrorTranslationFiles",
    "PassportElementErrorUnspecified",
    "Game",
    "CallbackGame",
    "GameHighScore"
]



from dataclasses import dataclass

from telebox.errors import TeleboxError
from telebox.dispatcher.typing import Event


@dataclass
class ContextError(TeleboxError):
    """Class for context errors."""


@dataclass
class InvalidEventError(ContextError):
    event: Event



from __future__ import annotations
from typing import TYPE_CHECKING

from telebox.dispatcher.filters.error_filter import AbstractErrorFilter
if TYPE_CHECKING:
    from telebox.dispatcher.dispatcher import Event


class ErrorNoneFilter(AbstractErrorFilter):

    def get_value(self, error: Exception, event: Event) -> bool:
        return True

    def check_value(self, value: bool) -> bool:
        return value



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type
from telebox.bot.types.types.user import User
from telebox.bot.types.types.message import Message


@dataclass(unsafe_hash=True)
class CallbackQuery(Type):
    id: str
    from_: User
    chat_instance: str
    message: Optional[Message] = None
    inline_message_id: Optional[str] = None
    data: Optional[str] = None
    game_short_name: Optional[str] = None

    @property
    def chat_type(self) -> Optional[str]:
        return self.message.chat.type if self.message is not None else None

    @property
    def chat_id(self) -> Optional[int]:
        return self.message.chat.id if self.message is not None else None

    @property
    def user_id(self) -> int:
        return self.from_.id

    @property
    def message_id(self) -> Optional[int]:
        return self.message.message_id if self.message is not None else None



from dataclasses import dataclass

from telebox.errors import TeleboxError


@dataclass
class TaskExecutorError(TeleboxError):
    """Class for task executor errors."""


@dataclass
class PendingTaskNotFoundError(TaskExecutorError):
    id: str



from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.bot.types.types.message import Message


class NewChatMemberIDFilter(AbstractEventFilter):

    def __init__(self, *ids: int):
        self._ids = set(ids)

    def get_event_types(self) -> set[EventType]:
        return {EventType.MESSAGE}

    def get_value(self, event: Message) -> set[int]:
        ids = set()

        if event.new_chat_members is not None:
            for i in event.new_chat_members:
                ids.add(i.id)

        return ids

    def check_value(self, value: set[int]) -> bool:
        return any(i in self._ids for i in value)



DATA = "data"
FRONT_SIDE = "front_side"
REVERSE_SIDE = "reverse_side"
SELFIE = "selfie"
FILE = "file"
FILES = "files"
TRANSLATION_FILE = "translation_file"
TRANSLATION_FILES = "translation_files"
UNSPECIFIED = "unspecified"



UNITED_ARAB_EMIRATES_DIRHAM = "AED"
AFGHAN_AFGHANI = "AFN"
ALBANIAN_LEK = "ALL"
ARMENIAN_DRAM = "AMD"
ARGENTINE_PESO = "ARS"
AUSTRALIAN_DOLLAR = "AUD"
AZERBAIJANI_MANAT = "AZN"
BOSNIA_AND_HERZEGOVINA_CONVERTIBLE_MARK = "BAM"
BANGLADESHI_TAKA = "BDT"
BULGARIAN_LEV = "BGN"
BRUNEI_DOLLAR = "BND"
BOLIVIAN_BOLIVIANO = "BOB"
BRAZILIAN_REAL = "BRL"
CANADIAN_DOLLAR = "CAD"
SWISS_FRANC = "CHF"
CHILEAN_PESO = "CLP"
CHINESE_RENMINBI_YUAN = "CNY"
COLOMBIAN_PESO = "COP"
COSTA_RICAN_COLON = "CRC"
CZECH_KORUNA = "CZK"
DANISH_KRONE = "DKK"
DOMINICAN_PESO = "DOP"
ALGERIAN_DINAR = "DZD"
EGYPTIAN_POUND = "EGP"
EURO = "EUR"
BRITISH_POUND = "GBP"
GEORGIAN_LARI = "GEL"
GUATEMALAN_QUETZAL = "GTQ"
HONG_KONG_DOLLAR = "HKD"
HONDURAN_LEMPIRA = "HNL"
CROATIAN_KUNA = "HRK"
HUNGARIAN_FORINT = "HUF"
INDONESIAN_RUPIAH = "IDR"
ISRAELI_NEW_SHEQEL = "ILS"
INDIAN_RUPEE = "INR"
ICELANDIC_KRONA = "ISK"
JAMAICAN_DOLLAR = "JMD"
JAPANESE_YEN = "JPY"
KENYAN_SHILLING = "KES"
KYRGYZSTANI_SOM = "KGS"
SOUTH_KOREAN_WON = "KRW"
KAZAKHSTANI_TENGE = "KZT"
LEBANESE_POUND = "LBP"
SRI_LANKAN_RUPEE = "LKR"
MOROCCAN_DIRHAM = "MAD"
MOLDOVAN_LEU = "MDL"
MONGOLIAN_TOGROG = "MNT"
MAURITIAN_RUPEE = "MUR"
MALDIVIAN_RUFIYAA = "MVR"
MEXICAN_PESO = "MXN"
MALAYSIAN_RINGGIT = "MYR"
MOZAMBICAN_METICAL = "MZN"
NIGERIAN_NAIRA = "NGN"
NICARAGUAN_CORDOBA = "NIO"
NORWEGIAN_KRONE = "NOK"
NEPALESE_RUPEE = "NPR"
NEW_ZEALAND_DOLLAR = "NZD"
PANAMANIAN_BALBOA = "PAB"
PERUVIAN_NUEVO_SOL = "PEN"
PHILIPPINE_PESO = "PHP"
PAKISTANI_RUPEE = "PKR"
POLISH_ZLOTY = "PLN"
PARAGUAYAN_GUARANI = "PYG"
QATARI_RIYAL = "QAR"
ROMANIAN_LEU = "RON"
SERBIAN_DINAR = "RSD"
RUSSIAN_RUBLE = "RUB"
SAUDI_RIYAL = "SAR"
SWEDISH_KRONA = "SEK"
SINGAPORE_DOLLAR = "SGD"
THAI_BAHT = "THB"
TAJIKISTANI_SOMONI = "TJS"
TURKISH_LIRA = "TRY"
TRINIDAD_AND_TOBAGO_DOLLAR = "TTD"
NEW_TAIWAN_DOLLAR = "TWD"
TANZANIAN_SHILLING = "TZS"
UKRAINIAN_HRYVNIA = "UAH"
UGANDAN_SHILLING = "UGX"
UNITED_STATES_DOLLAR = "USD"
URUGUAYAN_PESO = "UYU"
UZBEKISTANI_SOM = "UZS"
VIETNAMESE_DONG = "VND"
YEMENI_RIAL = "YER"
SOUTH_AFRICAN_RAND = "ZAR"



from typing import Optional

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.bot.types.types.message import Message


class LeftChatMemberIDFilter(AbstractEventFilter):

    def __init__(self, *ids: int):
        self._ids = set(ids)

    def get_event_types(self) -> set[EventType]:
        return {EventType.MESSAGE}

    def get_value(self, event: Message) -> Optional[int]:
        user = event.left_chat_member

        if user is not None:
            return user.id

    def check_value(self, value: Optional[int]) -> bool:
        return value in self._ids



from dataclasses import dataclass

from telebox.bot.types.type import Type
from telebox.bot.consts import bot_command_scope_types


@dataclass(unsafe_hash=True)
class BotCommandScopeAllGroupChats(Type):
    type: str = bot_command_scope_types.ALL_GROUP_CHATS



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type
from telebox.bot.types.types.web_app_info import WebAppInfo
from telebox.bot.types.types.login_url import LoginUrl
from telebox.bot.types.types.callback_game import CallbackGame


@dataclass(unsafe_hash=True)
class InlineKeyboardButton(Type):
    text: str
    url: Optional[str] = None
    callback_data: Optional[str] = None
    web_app: Optional[WebAppInfo] = None
    login_url: Optional[LoginUrl] = None
    switch_inline_query: Optional[str] = None
    switch_inline_query_current_chat: Optional[str] = None
    callback_game: Optional[CallbackGame] = None
    pay: Optional[bool] = None



from dataclasses import dataclass

from telebox.bot.types.type import Type
from telebox.bot.types.types.user import User


@dataclass(unsafe_hash=True)
class GameHighScore(Type):
    position: int
    user: User
    score: int



from __future__ import annotations
from typing import Optional, Union, Any, TYPE_CHECKING

from telebox.state_machine.state import State
from telebox.state_machine.storages.storage import AbstractStateStorage
from telebox.state_machine.manager import StateManager
from telebox.state_machine.transition_scheme import TransitionScheme
from telebox.state_machine.magazine import StateMagazine
from telebox.state_machine.errors import (
    DestinationStateNotFoundError,
    NextStateNotFoundError,
    PreviousStateNotFoundError
)
from telebox.dispatcher.handlers.event import AbstractEventHandler
if TYPE_CHECKING:
    from telebox.dispatcher.typing import Event


TransitionDict = dict[State, dict[AbstractEventHandler, Union[State, dict[str, State]]]]


class StateMachine:

    def __init__(self, initial_state: State, storage: AbstractStateStorage):
        self._state_manager = StateManager(initial_state, storage)
        self._transition_scheme = TransitionScheme()

    @property
    def initial_state(self) -> State:
        return self._state_manager.initial_state

    @property
    def states(self) -> set[State]:
        return self._state_manager.states

    def add_state(self, state: State) -> None:
        self._state_manager.add_state(state)

    def add_transition(
        self,
        source_state: State,
        destination_state: State,
        handler: AbstractEventHandler,
        direction: Optional[str] = None
    ) -> None:
        self._transition_scheme.add_transition(
            source_state=source_state,
            destination_state=destination_state,
            handler=handler,
            direction=direction
        )

        for i in (source_state, destination_state):
            if not self.check_state(i):
                self.add_state(i)

    def add_transitions(self, transitions: TransitionDict) -> None:
        for source_state in transitions:
            for handler, destination_item in transitions[source_state].items():
                if isinstance(destination_item, dict):
                    for direction, destination_state in destination_item.items():
                        self.add_transition(source_state, destination_state, handler, direction)
                else:
                    self.add_transition(source_state, destination_item, handler)

    def check_state(self, state: State) -> bool:
        return self._state_manager.check_state(state)

    def check_transition(
        self,
        source_state: State,
        destination_state: State,
        handler: AbstractEventHandler,
        direction: Optional[str] = None
    ) -> bool:
        return self._transition_scheme.check_transition(
            source_state,
            destination_state,
            handler,
            direction
        )

    def get_state(self, *, chat_id: int, user_id: Optional[int] = None) -> State:
        magazine = self._state_manager.load_magazine(chat_id=chat_id, user_id=user_id)

        return self._state_manager.get_state(magazine.current_state)

    def set_next_state(
        self,
        event: Event,
        handler: AbstractEventHandler,
        direction: Optional[str] = None,
        data: Any = None,
        *,
        chat_id: int,
        user_id: Optional[int] = None
    ) -> None:
        magazine = self._state_manager.load_magazine(chat_id=chat_id, user_id=user_id)
        current_state = self._state_manager.get_state(magazine.current_state)

        try:
            next_state = self._transition_scheme.get_destination_state(
                source_state=current_state,
                handler=handler,
                direction=direction
            )
        except DestinationStateNotFoundError as error:
            raise NextStateNotFoundError(
                "Transition to a next state has not been set "
                "({source_state=}, {handler=}, {direction=})!",
                source_state=error.source_state,
                handler=error.handler,
                direction=error.direction
            ) from None

        self._process_transition(
            event=event,
            magazine=magazine,
            source_state=current_state,
            destination_state=next_state,
            data=data,
            chat_id=chat_id,
            user_id=user_id
        )

    def set_previous_state(
        self,
        event: Event,
        data: Any = None,
        *,
        chat_id: int,
        user_id: Optional[int] = None
    ) -> None:
        magazine = self._state_manager.load_magazine(chat_id=chat_id, user_id=user_id)
        current_state = self._state_manager.get_state(magazine.current_state)

        if magazine.previous_state is None:
            raise PreviousStateNotFoundError(
                "A previous state cannot be found because the current "
                "state is the initial state!",
                current_state=current_state
            )

        previous_state = self._state_manager.get_state(magazine.previous_state)
        self._process_transition(
            event=event,
            magazine=magazine,
            source_state=current_state,
            destination_state=previous_state,
            data=data,
            chat_id=chat_id,
            user_id=user_id
        )

    def set_state(
        self,
        state: State,
        event: Event,
        data: Any = None,
        *,
        chat_id: int,
        user_id: Optional[int] = None
    ) -> None:
        magazine = self._state_manager.load_magazine(chat_id=chat_id, user_id=user_id)
        current_state = self._state_manager.get_state(magazine.current_state)
        self._process_transition(
            event=event,
            magazine=magazine,
            source_state=current_state,
            destination_state=state,
            data=data,
            chat_id=chat_id,
            user_id=user_id
        )

    def reset_state(
        self,
        event: Event,
        data: Any = None,
        *,
        chat_id: int,
        user_id: Optional[int] = None,
        with_exit: bool = True
    ) -> None:
        state = self.get_state(chat_id=chat_id, user_id=user_id)

        if with_exit:
            state.process_exit(event, data)

        state.process_enter(event, data)

    def _process_transition(
        self,
        event: Event,
        magazine: StateMagazine,
        source_state: State,
        destination_state: State,
        data: Any = None,
        *,
        chat_id: int,
        user_id: int
    ) -> None:
        source_state.process_exit(event, data)
        destination_state.process_enter(event, data)
        magazine.set_state(str(destination_state))
        self._state_manager.save_magazine(magazine, chat_id=chat_id, user_id=user_id)



from dataclasses import dataclass
from typing import Optional

from telebox.bot.types.type import Type


@dataclass(unsafe_hash=True)
class LoginUrl(Type):
    url: str
    forward_text: Optional[str] = None
    bot_username: Optional[str] = None
    request_write_access: Optional[bool] = None



from __future__ import annotations
from typing import TYPE_CHECKING

from telebox.dispatcher.filters.error_filter import AbstractErrorFilter
if TYPE_CHECKING:
    from telebox.dispatcher.typing import Event


class ErrorTextFilter(AbstractErrorFilter):

    def __init__(self, *texts: str, full_match: bool = True, ignore_case: bool = False):
        if not texts:
            raise ValueError("No texts!")

        self._texts = {i.lower() for i in texts} if ignore_case else set(texts)
        self._full_match = full_match
        self._ignore_case = ignore_case

    def get_value(self, error: Exception, event: Event) -> str:
        return str(error)

    def check_value(self, value: str) -> bool:
        if self._ignore_case:
            value = value.lower()

        if self._full_match:
            return value in self._texts
        else:
            return any(i in value for i in self._texts)



from dataclasses import dataclass

from telebox.bot.types.type import Type
from telebox.bot.consts import passport_element_error_sources


@dataclass(unsafe_hash=True)
class PassportElementErrorReverseSide(Type):
    type: str
    file_hash: str
    message: str
    source: str = passport_element_error_sources.REVERSE_SIDE



from __future__ import annotations
from typing import Pattern, TYPE_CHECKING

from telebox.dispatcher.filters.error_filter import AbstractErrorFilter
if TYPE_CHECKING:
    from telebox.dispatcher.typing import Event


class ErrorRegExpTextFilter(AbstractErrorFilter):

    def __init__(self, *patterns: Pattern, full_match: bool = True):
        if not patterns:
            raise ValueError("No patterns!")

        self._patterns = set(patterns)
        self._full_match = full_match

    def get_value(self, error: Exception, event: Event) -> str:
        return str(error)

    def check_value(self, value: str) -> bool:
        if self._full_match:
            return any(i.fullmatch(value) is not None for i in self._patterns)
        else:
            return any(i.match(value) is not None for i in self._patterns)



from .deep_links import (
    get_username_link,
    get_phone_number_link,
    get_chat_invite_link,
    get_message_public_link,
    get_message_private_link,
    get_share_link,
    get_video_chat_link,
    get_livestream_link,
    get_voice_chat_link,
    get_add_stickers_link,
    get_add_emoji_link,
    get_mtproxy_link,
    get_socks5_proxy_link,
    get_add_theme_link,
    get_bot_link,
    get_group_bot_link,
    get_channel_bot_link,
    get_game_link,
    get_user_link
)
from .formatting import get_escaped_html_text, get_escaped_markdown_text
from .users import get_full_name


__all__ = [
    "get_username_link",
    "get_phone_number_link",
    "get_chat_invite_link",
    "get_message_public_link",
    "get_message_private_link",
    "get_share_link",
    "get_video_chat_link",
    "get_livestream_link",
    "get_voice_chat_link",
    "get_add_stickers_link",
    "get_add_emoji_link",
    "get_mtproxy_link",
    "get_socks5_proxy_link",
    "get_add_theme_link",
    "get_bot_link",
    "get_group_bot_link",
    "get_channel_bot_link",
    "get_game_link",
    "get_user_link",
    "get_escaped_html_text",
    "get_escaped_markdown_text",
    "get_full_name"
]



from typing import Optional, Any

from telebox.dispatcher.filters.event_filter import AbstractEventFilter
from telebox.dispatcher.enums.event_type import EventType
from telebox.utils.callback_data_builders.builder import AbstractCallbackDataBuilder
from telebox.bot.types.types.callback_query import CallbackQuery


class CallbackDataFilter(AbstractEventFilter):

    def __init__(self, *keys: Any, builder: AbstractCallbackDataBuilder):
        self._keys = set(keys)
        self._builder = builder

    def get_event_types(self) -> set[EventType]:
        return {EventType.CALLBACK_QUERY}

    def get_value(self, event: CallbackQuery) -> Optional[str]:
        if event.data is not None:
            key, _ = self._builder.get_parsed_data(event.data)

            return key

    def check_value(self, value: Optional[str]) -> bool:
        return value in self._keys if self._keys else value is not None



from dataclasses import dataclass
from typing import Optional, Any
from http import HTTPStatus

from telebox.errors import TeleboxError
from telebox.bot.types.types.response_parameters import ResponseParameters


@dataclass
class BotError(TeleboxError):
    """Class for bot errors."""


@dataclass
class UnknownUpdateContentError(BotError):
    """Class for unknown update content error."""


@dataclass
class UnknownMessageContentError(BotError):
    """Class for unknown message content error."""


@dataclass
class RequestError(BotError):
    """Class for Telegram Bot API request error."""
    DEFAULT_TEMPLATE = "\n".join((
        "A request to the Telegram Bot API was unsuccessful!",
        "├─ Method: {method!r}",
        "├─ Parameters: {parameters!r}",
        "├─ Status code: {status_code!r}",
        "└─ Description: {description!r}"
    ))
    method: str
    parameters: dict[str, Any]
    status_code: int
    description: str


@dataclass
class BadRequestError(RequestError):
    """Error class for 400 status code."""


@dataclass
class ChatNotFoundError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: chat not found',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class UserNotFoundError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: user not found',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageIsNotModifiedError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message is not modified',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class ChatDescriptionIsNotModifiedError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: chat description is not modified',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageToForwardNotFoundError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message to forward not found',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageToDeleteNotFoundError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message to delete not found',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageToEditNotFoundError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message to edit not found',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class PollAlreadyClosedError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: poll has already been closed',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class PollMustHaveMoreOptionsError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: poll must have at least 2 option',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class PollCannotHaveMoreOptionsError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: poll can't have more than 10 options',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class PollOptionsMustBeNonEmptyError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: poll options must be non-empty',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class PollQuestionMustBeNonEmptyError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: poll question must be non-empty',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class PollOptionsLengthIsTooLongError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: poll options length must not exceed 100',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class PollQuestionLengthIsTooLongError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: poll question length must not exceed 255',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageWithPollToStopNotFoundError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message with poll to stop not found',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageIsNotPollError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message is not a poll',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class InvalidMessageIDError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: MESSAGE_ID_INVALID',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class PrivateChannelError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: CHANNEL_PRIVATE',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class WriteToChatForbiddenError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: CHAT_WRITE_FORBIDDEN',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageToPinNotFoundError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message to pin not found',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class NoRightsToManagePinnedMessagesError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: not enough rights to manage pinned messages in the chat',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class NoRightsToExportChatInviteLinkError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: not enough rights to export chat invite link',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class NoRightsToRestrictUnrestrictChatMemberError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: not enough rights to restrict/unrestrict chat member',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class NoRightsToSendMessageError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: have no rights to send a message',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class RepliedMessageNotFoundError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: replied message not found',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageIDIsNotSpecifiedError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message identifier is not specified',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageTextIsEmptyError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message text is empty',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageCannotBeEditedError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message can't be edited',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageCannotBeDeletedError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message can't be deleted',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class GroupWasUpgradedToSupergroupError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: group chat was upgraded to a supergroup chat',
        'error_code': 400,
        'ok': False,
        'parameters': {
            'migrate_to_chat_id': -100123456789
        }
    }
    """
    ALL_REPLACEMENT_FIELDS_IS_REQUIRED = False
    migrate_to_chat_id: int


@dataclass
class MessageIsTooLongError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: message is too long',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class QueryIsTooOldOrInvalidIDError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: query is too old and response timeout expired or query ID is invalid',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class InvalidButtonURLError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: BUTTON_URL_INVALID',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class InvalidButtonDataError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: BUTTON_DATA_INVALID',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class InlineKeyboardButtonParsingError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: can't parse inline keyboard button: Text buttons are unallowed in the inline keyboard',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class WrongFileIDError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: wrong file id',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class GroupIsDeactivatedError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: group is deactivated',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class PhotoShouldBeUploadedAsInputFileError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: photo should be uploaded as an InputFile',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class InvalidStickerSetError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: STICKERSET_INVALID',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class TooMuchMessagesToSendAsAlbumError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: too much messages to send as an album',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class DemoteChatCreatorError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: can't demote chat creator',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class SelfRestrictError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: can't restrict self',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class WebhookURLMustBeHTTPSError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: bad webhook: HTTPS url must be provided for webhook',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class URLParsingError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: can't parse URL',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class BadWebhookPortError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: bad webhook: Webhook can be set up only on ports 80, 88, 443 or 8443',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class UnknownWebhookHostError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: bad webhook: Failed to resolve host: Name or service not known',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class MessageEntitiesParsingError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: can't parse entities',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class WrongHTTPURLError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: wrong HTTP URL',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class InvalidFileIDError(BadRequestError):
    """Error class for this response:
    {
        'description': 'Bad Request: invalid file id',
        'error_code': 400,
        'ok': False
    }
    """


@dataclass
class UnauthorizedError(RequestError):
    """Error class for 401 status code."""


@dataclass
class ForbiddenError(RequestError):
    """Error class for 403 status code."""


@dataclass
class BotWasBlockedByUserError(ForbiddenError):
    """Error class for this response:
    {
        'description': 'Forbidden: bot was blocked by the user',
        'error_code': 403,
        'ok': False
    }
    """


@dataclass
class UserIsDeactivatedError(ForbiddenError):
    """Error class for this response:
    {
        'description': 'Forbidden: user is deactivated',
        'error_code': 403,
        'ok': False
    }
    """


@dataclass
class BotWasKickedFromGroupError(ForbiddenError):
    """Error class for this response:
    {
        'description': 'Forbidden: bot was kicked from the group chat',
        'error_code': 403,
        'ok': False
    }
    """


@dataclass
class BotWasKickedFromSupergroupError(ForbiddenError):
    """Error class for this response:
    {
        'description': 'Forbidden: bot was kicked from the supergroup chat',
        'error_code': 403,
        'ok': False
    }
    """


@dataclass
class BotCannotInitiateConversationWithUserError(ForbiddenError):
    """Error class for this response:
    {
        'description': 'Forbidden: bot can't initiate conversation with a user',
        'error_code': 403,
        'ok': False
    }
    """


@dataclass
class DeleteMessageForbiddenError(ForbiddenError):
    """Error class for this response:
    {
        'description': 'Forbidden: MESSAGE_DELETE_FORBIDDEN',
        'error_code': 403,
        'ok': False
    }
    """


@dataclass
class BotIsNotSupergroupMemberError(ForbiddenError):
    """Error class for this response:
    {
        'description': 'Forbidden: bot is not a member of the supergroup chat',
        'error_code': 403,
        'ok': False
    }
    """


@dataclass
class BotIsNotChannelMemberError(ForbiddenError):
    """Error class for this response:
    {
        'description': 'Forbidden: bot is not a member of the channel chat',
        'error_code': 403,
        'ok': False
    }
    """


@dataclass
class BotCannotSendMessagesToBotsError(ForbiddenError):
    """Error class for this response:
    {
        'description': 'Forbidden: bot can't send messages to bots',
        'error_code': 403,
        'ok': False
    }
    """


@dataclass
class ConflictError(RequestError):
    """Error class for 409 status code."""


@dataclass
class MultipleGetUpdatesError(ConflictError):
    """Error class for this response:
    {
        'description': 'Conflict: terminated by other getUpdates request; make sure that only one bot instance is running',
        'error_code': 409,
        'ok': False
    }
    """


@dataclass
class CannotUseGetUpdatesWhileWebhookIsActiveError(ConflictError):
    """Error class for this response:
    {
        'description': 'Conflict: can't use getUpdates method while webhook is active; use deleteWebhook to delete the webhook first',
        'error_code': 409,
        'ok': False
    }
    """


@dataclass
class RequestEntityTooLargeError(RequestError):
    """Error class for 413 status code."""


@dataclass
class TooManyRequestsError(RequestError):
    """Error class for 429 status code."""


@dataclass
class RetryAfterError(TooManyRequestsError):
    """Error class for this response:
    {
        'description': 'Too Many Requests: retry after 42',
        'error_code': 429,
        'ok': False,
        'parameters': {
            'retry_after': 42
        }
    }
    """
    ALL_REPLACEMENT_FIELDS_IS_REQUIRED = False
    retry_after: int


@dataclass
class InternalServerError(RequestError):
    """Error class for 500+ status code."""


@dataclass
class ServerIsRestartingError(InternalServerError):
    """Error class for this response:
    {
        'description': 'Internal Server Error: restart',
        'error_code': 500,
        'ok': False
    }
    """


@dataclass
class BadGatewayError(InternalServerError):
    """Error class for this response:
        {
            'description': 'Bad Gateway',
            'error_code': 502,
            'ok': False
        }
    """


def get_request_error(
    method: str,
    parameters: dict[str, Any],
    status_code: int,
    description: str,
    *,
    response_parameters: Optional[ResponseParameters] = None
) -> RequestError:
    kwargs = {
        "method": method,
        "parameters": {
            name: value
            for name, value in parameters.items()
            if value is not None
        },
        "status_code": status_code,
        "description": description
    }

    if status_code == HTTPStatus.BAD_REQUEST:
        return _get_bad_request_error(kwargs, response_parameters)
    elif status_code == HTTPStatus.UNAUTHORIZED:
        return _get_unauthorized_error(kwargs)
    elif status_code == HTTPStatus.FORBIDDEN:
        return _get_forbidden_error(kwargs)
    elif status_code == HTTPStatus.CONFLICT:
        return _get_conflict_error(kwargs)
    elif status_code == HTTPStatus.REQUEST_ENTITY_TOO_LARGE:
        return _get_request_entity_too_large_error(kwargs)
    elif status_code == HTTPStatus.TOO_MANY_REQUESTS:
        return _get_too_many_requests_error(kwargs, response_parameters)
    elif status_code >= HTTPStatus.INTERNAL_SERVER_ERROR:
        return _get_internal_server_error(kwargs)

    return RequestError(None, **kwargs)


def _get_bad_request_error(
    kwargs: dict[str, Any],
    response_parameters: ResponseParameters
) -> BadRequestError:
    lowered_description = kwargs["description"].lower()

    for message, error_type in (
        ("chat not found", ChatNotFoundError),
        ("user not found", UserNotFoundError),
        ("message is not modified", MessageIsNotModifiedError),
        ("chat description is not modified", ChatDescriptionIsNotModifiedError),
        ("message to forward not found", MessageToForwardNotFoundError),
        ("message to delete not found", MessageToDeleteNotFoundError),
        ("message to edit not found", MessageToEditNotFoundError),
        ("poll has already been closed", PollAlreadyClosedError),
        ("poll must have at least", PollMustHaveMoreOptionsError),
        ("poll can't have more than", PollCannotHaveMoreOptionsError),
        ("poll options must be non-empty", PollOptionsMustBeNonEmptyError),
        ("poll question must be non-empty", PollQuestionMustBeNonEmptyError),
        ("poll options length must not exceed", PollOptionsLengthIsTooLongError),
        ("poll question length must not exceed", PollQuestionLengthIsTooLongError),
        ("message with poll to stop not found", MessageWithPollToStopNotFoundError),
        ("message is not a poll", MessageIsNotPollError),
        ("message_id_invalid", InvalidMessageIDError),
        ("channel_private", PrivateChannelError),
        ("chat_write_forbidden", WriteToChatForbiddenError),
        ("message to pin not found", MessageToPinNotFoundError),
        (
            "not enough rights to manage pinned messages in the chat",
            NoRightsToManagePinnedMessagesError
        ),
        ("not enough rights to export chat invite link", NoRightsToExportChatInviteLinkError),
        (
            "not enough rights to restrict/unrestrict chat member",
            NoRightsToRestrictUnrestrictChatMemberError
        ),
        ("have no rights to send a message", NoRightsToSendMessageError),
        ("replied message not found", RepliedMessageNotFoundError),
        ("message identifier is not specified", MessageIDIsNotSpecifiedError),
        ("message text is empty", MessageTextIsEmptyError),
        ("message can't be edited", MessageCannotBeEditedError),
        ("message can't be deleted", MessageCannotBeDeletedError),
        ("group chat was upgraded to a supergroup chat", GroupWasUpgradedToSupergroupError),
        ("message is too long", MessageIsTooLongError),
        (
            "query is too old and response timeout expired or query id is invalid",
            QueryIsTooOldOrInvalidIDError
        ),
        ("button_url_invalid", InvalidButtonURLError),
        ("button_data_invalid", InvalidButtonDataError),
        ("can't parse inline keyboard button", InlineKeyboardButtonParsingError),
        ("wrong file id", WrongFileIDError),
        ("group is deactivated", GroupIsDeactivatedError),
        ("photo should be uploaded as an inputfile", PhotoShouldBeUploadedAsInputFileError),
        ("stickerset_invalid", InvalidStickerSetError),
        ("too much messages to send as an album", TooMuchMessagesToSendAsAlbumError),
        ("can't demote chat creator", DemoteChatCreatorError),
        ("can't restrict self", SelfRestrictError),
        ("https url must be provided for webhook", WebhookURLMustBeHTTPSError),
        ("can't parse url", URLParsingError),
        ("webhook can be set up only on ports", BadWebhookPortError),
        ("failed to resolve host: name or service not known", UnknownWebhookHostError),
        ("can't parse entities", MessageEntitiesParsingError),
        ("wrong http url", WrongHTTPURLError),
        ("invalid file id", InvalidFileIDError)
    ):
        if message in lowered_description:
            return error_type(None, **kwargs)

    if "group chat was upgraded to a supergroup chat" in lowered_description:
        return GroupWasUpgradedToSupergroupError(
            None,
            **kwargs,
            migrate_to_chat_id=response_parameters.migrate_to_chat_id
        )

    return BadRequestError(None, **kwargs)


def _get_unauthorized_error(kwargs: dict[str, Any]) -> UnauthorizedError:
    return UnauthorizedError(None, **kwargs)


def _get_forbidden_error(kwargs: dict[str, Any]) -> ForbiddenError:
    lowered_description = kwargs["description"].lower()

    for message, error_type in (
        ("bot was blocked by the user", BotWasBlockedByUserError),
        ("user is deactivated", UserIsDeactivatedError),
        ("bot was kicked from the group chat", BotWasKickedFromGroupError),
        ("bot was kicked from the supergroup chat", BotWasKickedFromSupergroupError),
        (
            "bot can't initiate conversation with a user",
            BotCannotInitiateConversationWithUserError
        ),
        ("message_delete_forbidden", DeleteMessageForbiddenError),
        ("bot is not a member of the supergroup chat", BotIsNotSupergroupMemberError),
        ("bot is not a member of the channel chat", BotIsNotChannelMemberError),
        ("bot can't send messages to bots", BotCannotSendMessagesToBotsError)
    ):
        if message in lowered_description:
            return error_type(None, **kwargs)

    return ForbiddenError(None, **kwargs)


def _get_conflict_error(kwargs: dict[str, Any]) -> ConflictError:
    lowered_description = kwargs["description"].lower()

    for message, error_type in (
        ("terminated by other getupdates request", MultipleGetUpdatesError),
        (
            "can't use getupdates method while webhook is active",
            CannotUseGetUpdatesWhileWebhookIsActiveError
        )
    ):
        if message in lowered_description:
            return error_type(None, **kwargs)

    return ConflictError(None, **kwargs)


def _get_request_entity_too_large_error(kwargs: dict[str, Any]) -> RequestEntityTooLargeError:
    return RequestEntityTooLargeError(None, **kwargs)


def _get_too_many_requests_error(
    kwargs: dict[str, Any],
    response_parameters: ResponseParameters
) -> TooManyRequestsError:
    lowered_description = kwargs["description"].lower()

    if "retry after" in lowered_description:
        return RetryAfterError(None, **kwargs, retry_after=response_parameters.retry_after)

    return TooManyRequestsError(None, **kwargs)


def _get_internal_server_error(kwargs: dict[str, Any]) -> InternalServerError:
    lowered_description = kwargs["description"].lower()

    for message, error_type in (
        ("restart", ServerIsRestartingError),
        ("bad gateway", BadGatewayError)
    ):
        if message in lowered_description:
            return error_type(None, **kwargs)

    return InternalServerError(None, **kwargs)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from telebox.bot.types.types.message import Message
from telebox.dispatcher.enums.media_group_content_type import MediaGroupContentType


@dataclass(unsafe_hash=True)
class MediaGroup:
    messages: list[Message]

    def __iter__(self):
        return iter(self.messages)

    def __len__(self):
        return len(self.messages)

    def __getitem__(self, item):
        return self.messages.__getitem__(item)

    @property
    def id(self) -> str:
        return self.messages[0].media_group_id

    @property
    def message_thread_id(self) -> Optional[int]:
        return self.messages[0].message_thread_id

    @property
    def content_types(self) -> set[MediaGroupContentType]:
        content_types = set()

        for i in self.messages:
            _, content_type = i.content
            content_type = MediaGroupContentType(content_type.value)
            content_types.add(content_type)

        return content_types

    @property
    def date(self) -> datetime:
        return self.messages[0].date

    @property
    def chat_type(self) -> str:
        return self.messages[0].chat_type

    @property
    def chat_id(self) -> int:
        return self.messages[0].chat_id

    @property
    def sender_chat_id(self) -> Optional[int]:
        return self.messages[0].sender_chat_id

    @property
    def user_id(self) -> Optional[int]:
        return self.messages[0].user_id

    @property
    def is_forwarded(self) -> bool:
        return self.messages[0].is_forwarded

    @property
    def is_reply(self) -> bool:
        return self.messages[0].is_reply



from .media_group import MediaGroup
from .rate_limiter import RateLimit, Limit
from .router import Router
from .events import get_event_chat_id, get_event_user_id


__all__ = [
    "MediaGroup",
    "RateLimit",
    "Limit",
    "Router",
    "get_event_chat_id",
    "get_event_user_id"
]
